
------------------------------------------------------------------------
--
-- AbsorbsMonitor
--
-- Copyright (C) 2010  Philipp Schmidt
--
-- This program is free software; you can redistribute it and/or
-- modify it under the terms of the GNU General Public License
-- as published by the Free Software Foundation; either version 2
-- of the License, or (at your option) any later version.
--
-- This program is distributed in the hope that it will be useful,
-- but WITHOUT ANY WARRANTY; without even the implied warranty of
-- MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
-- GNU General Public License for more details.
--
-- You should have received a copy of the GNU General Public License
-- along with this program; if not, write to:
--
-- Free Software Foundation, Inc.,
-- 51 Franklin Street, Fifth Floor,
-- Boston, MA  02110-1301, USA.
--
--
------------------------------------------------------------------------


local AM_Public = _G.AbsorbsMonitor;
local AM_Core = AM_Public.GetCoreObject();

local EffectType = AM_Public.EffectType;

local Scaling;
local playerScaling;
local privateScaling;

local playerGUID;
local playerClass;

local zoneModifier;

local AM_PushCharge = AM_Public.PushCharge;
local AM_PopCharge = AM_Public.PopCharge;

local AM_Unit_Stats = AM_Public.Unit_Stats;
local AM_Unit_Scaling = AM_Public.Unit_Scaling;
local AM_Unit_StatsAndScaling = AM_Public.Unit_StatsAndScaling;


-- #################################################################################################


-- No function in this file will be called before this
function AM_Core.EnableEffects()
	playerGUID = UnitGUID("player");
	_, playerClass = UnitClass("player");
	
	Scaling = AM_Public.Unit_ScalingAll();
	
	playerScaling = AM_Unit_Scaling(playerGUID);
	privateScaling = AM_Unit_Scaling(-1);
end

function AM_Core.SetZoneModifier(in_zoneModifier)
	zoneModifier = in_zoneModifier;
end


-- #################################################################################################


local wow_UnitHealthMax = UnitHealthMax;

local function UnitHealthMax(guid, name)
	-- Use original function if called with only one argument
	if(not name) then
		return wow_UnitHealthMax(guid);
	end

	local maxHealth = wow_UnitHealthMax(name);
	
	if(not maxHealth) then	
		if(UnitGUID("target") == guid) then
			maxHealth = wow_UnitHealthMax("target");
		elseif(UnitGUID("focus") == guid) then
			maxHealth = wow_UnitHealthMax("focus");
		end
	end	
	
	return maxHealth;
end


-- #################################################################################################


--- Generic Create function (only for documentary purposes)
-- @param	sourceGUID			guid of the originating unit
-- @param	sourceName			name of the originating unit
-- @param	destGUID			guid of the affected unit
-- @param	destName			name of the affected unit
-- @param	spellId				the spellId of this absorb effect
-- @param	destEffects			activeEffectsBySpell[destGUID]
-- @return	total value of this absorb effect
-- @return	quality (that is accuracy) of this value
local function generic_Create(sourceGUID, sourceName, destGUID, destName, spellId, destEffects)
end

local function generic_ConstantByTable_Create(sourceGUID, sourceName, destGUID, destName, spellId, destEffects)
	return AM_Public.Effects[spellId][6][spellId], 1.0;
end

-- Generic Create function for spells simply scaling with spellpower and a fixed
-- coefficient.
-- Expects at effect[6] a table indexed by spellId with the base values and at
-- effect[7] the spellpower coefficient
local function generic_SpellScalingByTable_Create(sourceGUID, sourceName, destGUID, destName, spellId, destEffects)	
	local effect = AM_Public.Effects[spellId];
	local _, sp, quality = AM_Unit_Stats(sourceGUID, 0.1);

	return floor(effect[6][spellId] + (sp * effect[7])), quality;
end

--- Generic Hit function suitable for most absorb effects
-- Note that this function is only responsible for determining the amount this
-- particular absorb effect will take, NOT to handle its consequences like updating
-- the data structures
-- @param	guid				guid of affected unit
-- @param	effectEntry			activeEffectsBySpell[guid][spellId]
-- @param	absorbedRemaining	absorb value left to be accounted for on this unit
-- @param	overkill			amount of damage done on top of the absorb
-- @param	spellSchool			spell school for this hit
-- @return	absorb value this effect can account fors
-- @return	whether this absorb was broken by this hit
local function generic_Hit(guid, effectEntry, absorbedRemaining, overkill, spellSchool)
	if(absorbedRemaining > effectEntry[3]) then
		-- dirty but efficient
		absorbedRemaining = effectEntry[3];
		overkill = 1;
	end

	return absorbedRemaining, (overkill == 0);
end


-- #################################################################################################


local deathknight_MS_Ranks = {[0] = 0, [1] = 0.08, [2] = 0.16, [3] = 0.25};

-- Public Scaling: { [MagicSuppression] }
local deathknight_defaultScaling = {0};


function deathknight_AntiMagicShell_Create(sourceGUID, sourceName, destGUID, destName, spellId, destEffects)
	local maxHealth = UnitHealthMax(destGUID, destName);
	
	if(not maxHealth) then	
		return 0, 0.0;
	end
	
	local sourceScaling, quality = AM_Unit_Scaling(sourceGUID, deathknight_defaultScaling, 0.5);
	
	return floor(maxHealth * 0.5), quality, (0.75 + sourceScaling[1]);
end

local function deathknight_AntiMagicShell_Hit(guid, effectEntry, absorbedRemaining, overkill, spellSchool) 
	-- TODO: what happens to mixed school attacks?
	if(spellSchool == SCHOOL_MASK_PHYSICAL) then 
		return 0, true;
	end
	
	local maxAbsorb = floor((absorbedRemaining + overkill) * effectEntry[7]);
	
	if(effectEntry[3] < maxAbsorb) then
		return effectEntry[3], false;
	else
		return maxAbsorb, true;
	end
end

local function deathknight_AntiMagicZone_Create(sourceGUID, sourceName, destGUID, destName, spellId, destEffects)
	local ap, _, quality = AM_Unit_Stats(sourceGUID, 0.3);

	return (10000 + 2 * ap), quality;
end

local function deathknight_AntiMagicZone_Hit(guid, effectEntry, absorbedRemaining, overkill, spellSchool)
	-- TODO: what happens to mixed school attacks?
	if(spellSchool == SCHOOL_MASK_PHYSICAL) then 
		return 0, true;
	end
	
	local maxAbsorb = floor((absorbedRemaining + overkill) * 0.75);
	
	if(effectEntry[3] < maxAbsorb) then
		return effectEntry[3], false;
	else
		return maxAbsorb, true;
	end
end

local function deathknight_OnTalentUpdate()
	-- Magic Suppression
	local _, _, _, _, t = GetTalentInfo(3, 18);
	
	playerScaling[1] = deathknight_MS_Ranks[t];
	
	AM_Public.ScheduleScalingBroadcast();
end

AM_Public.OnEnableClass.DEATHKNIGHT = function()
	AM_Public.OnTalentUpdate = deathknight_OnTalentUpdate;
		
	deathknight_OnTalentUpdate();
end


-- #################################################################################################


local function druid_SavageDefense_Create(sourceGUID, sourceName, destGUID, destName, spellId, destEffects)
	local ap, _, quality = AM_Unit_Stats(sourceGUID, 0.0);

	return floor(ap * 0.25), quality;
end

local function druid_SavageDefense_Hit(guid, effectEntry, absorbedRemaining, overkill, spellSchool) 
	-- TODO: what happens to mixed school attacks?
	if(spellSchool == SCHOOL_MASK_PHYSICAL) then 
		return min(effectEntry[3], absorbedRemaining), false;
	else
		return 0, true;
	end
end


-- #################################################################################################


-- Table for base values of
-- Fire Ward, Frost Ward, Ice Barrier, Mana Shield
-- TODO: base leveling increase
local mage_Absorb_Spells = {
	-- Fire Ward
	[543] = 165,
	[8457] = 290,
	[8458] = 470,
	[10223] = 675,
	[10225] = 875,
	[27218] = 1125,
	[43010] = 1950,
	
	-- Frost Ward
	[6143] = 165,
	[8461] = 290,
	[8462] = 470,
	[10177] = 675,
	[28609] = 875,
	[32796] = 1125,
	[43012] = 1950,
	
	-- Ice Barrier
	[11426] = 438,
	[13031] = 549,
	[13032] = 678,
	[13033] = 818,
	[27134] = 925,
	[33405] = 1075,
	[43038] = 2800,
	[43039] = 3300,
	
	-- Mana Shield
	[1463] = 120,
	[8494] = 210,
	[8495] = 300,
	[10191] = 390,
	[10192] = 480,
	[10193] = 570,
	[27131] = 715,
	[43019] = 1080,
	[43020] = 1330,
};

-- Public Scaling: { [GlyphOfIceBarrier] }
local mage_defaultScaling = {1.0};


-- No Downranking support here
local function mage_IceBarrier_Create(sourceGUID, sourceName, destGUID, destName, spellId, destEffects)
	local _, sp, quality1, sourceScaling, quality2 = AM_Unit_StatsAndScaling(sourceGUID, 0.3, mage_defaultScaling, 0.4);
	
	return floor((mage_Absorb_Spells[spellId] + (sp * 0.8053)) * sourceScaling[1]), math.min(quality1, quality2);
end

local function mage_FireWard_Hit(guid, effectEntry, absorbedRemaining, overkill, spellSchool)
	if(spellSchool ~= SCHOOL_MASK_FIRE) then
		return 0, true;
	end
	
	return generic_Hit(guid, effectEntry, absorbedRemaining, overkill, spellSchool);
end

local function mage_FrostWard_Hit(guid, effectEntry, absorbedRemaining, overkill, spellSchool)
	if(spellSchool ~= SCHOOL_MASK_FROST) then
		return 0, true;
	end
	
	return generic_Hit(guid, effectEntry, absorbedRemaining, overkill, spellSchool);
end

local function mage_OnGlyphUpdated()
	local glyphSpellId;
	
	playerScaling[1] = 1.0;

	for i = 1, 6 do
		_, _, glyphSpellId = GetGlyphSocketInfo(i);
		
		-- Glyph of Ice Barrier
		if(glyphSpellId and glyphSpellId == 63095) then
			playerScaling[1] = 1.3;
			
			break;
		end
	end
	 
	AM_Public.ScheduleScalingBroadcast();		
end

AM_Public.OnEnableClass.MAGE = function()
	AM_Public.OnGlyphUpdated = mage_OnGlyphUpdated;
	
	mage_OnGlyphUpdated();
end


-- #################################################################################################


-- Public Scaling: { [DivineGuardian] }
local paladin_defaultScaling = {1.0};


-- The base value is always 500
local function paladin_SacredShield_Create(sourceGUID, sourceName, destGUID, destName, spellId, destEffects)
	local _, sp, quality1, sourceScaling, quality2 = AM_Unit_StatsAndScaling(sourceGUID, 0.1, paladin_defaultScaling, 0.2);

	return floor((500 + (sp * 0.75)) * sourceScaling[1] * zoneModifier), math.min(quality1, quality2);
end

local function paladin_OnTalentUpdate()
	-- No need to do it before
	if(UnitLevel("player") < 80) then return; end

	-- Divine Guardian
	local _, _, _, _, t = GetTalentInfo(2, 9);
	
	playerScaling[1] = 1 + (t * 0.1);
	
	AM_Public.ScheduleScalingBroadcast();	
end

AM_Public.OnEnableClass.PALADIN = function()
	AM_Public.OnTalentUpdate = paladin_OnTalentUpdate;
	
	paladin_OnTalentUpdate();
end


-- #################################################################################################


PRIEST_DIVINEAEGIS_SPELLID = 47753;

-- [rank] = {spellId, level, baseValue, incValue}
local priest_PWS_Ranks = {
	[1] = {17, 6, 44, 4},
	[2] = {592, 12, 88, 6},
	[3] = {600, 18, 158, 8},
	[4] = {3747, 24, 234, 10},
	[5] = {6065, 30, 301, 11},
	[6] = {6066, 36, 381, 13},
	[7] = {10898, 42, 484, 15},
	[8] = {10899, 48, 605, 17},
	[9] = {10900, 54, 763, 19},
	[10] = {10901, 60, 942, 21},
	[11] = {25217, 65, 1125, 18},
	[12] = {25218, 70, 1265, 20},
	[13] = {48065, 75, 1920, 30},
	[14] = {48066, 80, 2230, 0}
};

-- Public Scaling:
--   Power Word: Shield: [spellId] = {base, spFactor}
--   Divine Aegis: [47753] = healFactor
local priest_defaultScaling = {[PRIEST_DIVINEAEGIS_SPELLID] = 0};

do 
	for k, v in pairs(priest_PWS_Ranks) do
		priest_defaultScaling[v[1]] = {v[3], 0.809};
	end
end

-- Private Scaling
--   Talents: "TwinDisc", "ImpPWS", "FocusedPower", "DivineAegis", "BorrowedTime", "SpiritualHealing"
--	 Gear: "4pcRaid9", "4pcRaid10"
--   Computed: base, sp, DA


local function priest_PowerWordShield_Create(sourceGUID, sourceName, destGUID, destName, spellId, destEffects)	
	local _, sp, quality1, sourceScaling, quality2 = AM_Unit_StatsAndScaling(sourceGUID, 0.1, priest_defaultScaling, 0.1);
	
	return floor((sourceScaling[spellId][1] + sp * sourceScaling[spellId][2]) * zoneModifier), math.min(quality1, quality2);
end

local function priest_DivineAegis_Create(sourceGUID, sourceName, destGUID, destName, spellId, destEffects)		
	local existing = 0;
	
	if(destEffects and destEffects[spellId]) then
		existing = destEffects[spellId][3];
	end

	local charge = AM_PopCharge(destGUID, spellId);
	
	if(charge == 0) then
		return existing, 0.0;
	end
	
	local destLevel = UnitLevel(destName);
	local quality = 1.0;		
	
	if(not destLevel) then
		destLevel = 80;
		quality = 0.4;
	end
		
	return min(destLevel * 125, existing + charge), quality;
end

-- I officially HATE Divine Aegis (and Val'anyr for that matter) from now. 
-- After extensive testing and parsing/filtering a few hours of combat log, I found the following facts:
-- 	* In MOST cases, every critical heal will trigger an AURA_APPLIED/AURA_REFRESHED event, even on multiple crits on
--    one penance and with both AURA events being triggered after all critical heals. In rare cases, there is only one...
--	* If on your side the aura get removed before the next one is applied (double penance crit, both critical heals
--    are first in your combat log), then AURA_APPLIED -> AURA_REMOVED by damage -> AURA_APPLIED for 2nd crit
--	* There are occasions where 2 Discipline priests apply two Divine Aegis auras. In general, the source of Divine Aegis
--    is completely fucked up in those cases. If two priests channel penance at the same time, both crit, you can never say
--    who will get the AURA_APPLIED event credited (yes, I found both the first-hitting priest as well as the second-hitting
--    priest there!) and who the following AURA_REFRESHED events
local function priest_DivineAegis_OnHealCrit(sourceGUID, sourceName, destGUID, destName, spellId, amount)	
	-- We can do a direct access to Scaling here, since the callback is only in place if it was present
	-- in the first place
	
	AM_PushCharge(destGUID, PRIEST_DIVINEAEGIS_SPELLID, floor(amount * Scaling[sourceGUID][PRIEST_DIVINEAEGIS_SPELLID]), 5.0);
end

local function priest_ApplyScaling(guid, level, baseFactor, spFactor, daFactor)
	local guidScaling;
	
	--AbsorbsMonitor:Print("Applying for "..guid.." ("..level..") with factors: "..baseFactor.." / "..spFactor.." / "..daFactor);

	if(not Scaling[guid]) then
		guidScaling = {};
		Scaling[guid] = guidScaling;
	else
		guidScaling = Scaling[guid];
	end
	
	guidScaling[PRIEST_DIVINEAEGIS_SPELLID] = daFactor;

	if(daFactor > 0) then
		AM_Public.AddCombatTrigger(guid, "OnHealCrit", priest_DivineAegis_OnHealCrit);
	else
		AM_Public.RemoveCombatTrigger(guid, "OnHealCrit", priest_DivineAegis_OnHealCrit);
	end
	
	local rankValue, rankSP;
	
	for k, v in pairs(priest_PWS_Ranks) do
		if(v[2] <= level) then
			if(level == 80) then
				rankValue = v[3] + v[4];
			else
				-- TODO
				rankValue = v[3];
			end
		
			-- Based on the assumption that the decrease in sp coefficient is linear,
			-- only tested for level 80 though so far
			-- Cataclysm will help us get rid of this crap again
			if(v[2] < (level - 5)) then
				if(level == 80) then
					rankSP = max(spFactor * ((v[2] * 0.045228921) - 2.381389768), 0);
				else
					rankSP = 0;
				end
			else
				rankSP = spFactor;
			end
			
			guidScaling[v[1]] = {rankValue * baseFactor, rankSP};
		end
	end
end

local function priest_UpdatePlayerScaling()
	privateScaling.base = ( (1.0 + (privateScaling["TwinDisc"] * 0.01) + (privateScaling["FocusedPower"] * 0.02) + (privateScaling["SpiritualHealing"] * 0.02)) * (1.0 + ((privateScaling["ImpPWS"] + privateScaling["4pcRaid10"]) * 0.05)) );
	
	local spFactor = 0.807;
	spFactor = spFactor + (privateScaling["BorrowedTime"] * 0.08);
	spFactor = spFactor * (1.0 + (privateScaling["TwinDisc"] * 0.01) + (privateScaling["FocusedPower"] * 0.02) + (privateScaling["SpiritualHealing"] * 0.02)) * (1.0 + ((privateScaling["ImpPWS"] + privateScaling["4pcRaid10"]) * 0.05));	
	privateScaling.sp = spFactor;	
	
	privateScaling.DA = (privateScaling["DivineAegis"] * 0.1) * (1 + (privateScaling["4pcRaid9"] * 0.03));

	priest_ApplyScaling(playerGUID, UnitLevel("player"), privateScaling.base, privateScaling.sp, privateScaling.DA);
	
	AM_Public.ScheduleScalingBroadcast();	
end

--[[
-- Since the tooltips only SOMETIMES include talents, scanning the spellbook is not very
-- reliable. This code may be useful for Cataclysm though, depending on how easy the scaling
-- formulas are
local function priest_ScanSpellbook()
	local PWS_Tooltip = {};
	PWS_Tooltip.t = CreateFrame("GameTooltip");
	PWS_Tooltip.t:SetOwner(UIParent, "ANCHOR_NONE");
	
	for i = 1, 4 do
		PWS_Tooltip["TooltipTextLeft" .. i] = PWS_Tooltip.t:CreateFontString();
		PWS_Tooltip["TooltipTextRight" .. i] = PWS_Tooltip.t:CreateFontString();
		PWS_Tooltip.t:AddFontStrings(PWS_Tooltip["TooltipTextLeft" .. i], PWS_Tooltip["TooltipTextRight" .. i]);
	end
		
	local spellName, spellRank, spellId;
	local i = 1;
	
	while true do
		spellName, spellRank = GetSpellName(i, BOOKTYPE_SPELL);
		
		if(not spellName) then break; end
		
		if(spellName == "Power Word: Shield") then
			spellRank = tonumber(spellRank:match("(%d+)"));			
			spellId = priest_PWS_Ranks[spellRank][1];
			
			if(not spellId) then
				AM_Core:Print("could not found spell id for rank "..spellRank);
				break;
			end						
			
			PWS_Tooltip.t:SetSpell(i, BOOKTYPE_SPELL);
			
			local _, _, value = string.find(PWS_Tooltip.TooltipTextLeft4:GetText(), "absorbing (%d+) damage");						
			
			AbsorbsMonitor:Print(spellRank.." - "..value);
			
			priest_PWS_Ranks[spellRank][3] = value;			
		end
		
		i = i + 1;
	end	
	
	PWS_Tooltip.t:SetParent(nil);
	PWS_Tooltip = nil;
end
]]

local function priest_ScanTalents()
	-- Twin Disciplines
	local _, _, _, _, t = GetTalentInfo(1, 2);
	privateScaling["TwinDisc"] = t;
	
	-- Improved Power Word: Shield
	local _, _, _, _, t = GetTalentInfo(1, 9);
	privateScaling["ImpPWS"] = t;
	
	-- Focused Power
	local _, _, _, _, t = GetTalentInfo(1, 16);
	privateScaling["FocusedPower"] = t;
	
	-- Divine Aegis
	local _, _, _, _, t = GetTalentInfo(1, 24);		
	privateScaling["DivineAegis"] = t;
	
	-- Borrowed Time
	local _, _, _, _, t = GetTalentInfo(1, 27);
	privateScaling["BorrowedTime"] = t;
	
	-- Spiritual Healing
	local _, _, _, _, t = GetTalentInfo(2, 16);
	privateScaling["SpiritualHealing"] = t;
end

local function priest_ScanEquipment()
	local n = 0;
	
	-- Crimson Acolyte Raiment's 4-piece Bonus
	if(IsEquippedItem(50765) or IsEquippedItem(51178) or IsEquippedItem(51261))		then n = 1; end
	if(IsEquippedItem(50767) or IsEquippedItem(51175) or IsEquippedItem(51264))		then n = n + 1; end
	if(IsEquippedItem(50768) or IsEquippedItem(51176) or IsEquippedItem(51263)) 	then n = n + 1; end
	if(IsEquippedItem(50766) or IsEquippedItem(51179) or IsEquippedItem(51260)) 	then n = n + 1; end
	if(IsEquippedItem(50769) or IsEquippedItem(51177) or IsEquippedItem(51262)) 	then n = n + 1; end
	
	if(n >= 4) then
		privateScaling["4pcRaid10"] = 1;
		
		 -- no way to have 4pcRaid9 now
		privateScaling["4pcRaid9"] = 0;		
		return;
	else
		privateScaling["4pcRaid10"] = 0;
	end
	
	n = 0;
	
	-- Velen's/Zabra's Raiment 4-piece Bonus
	
	-- UnitFactionGroup's first return value is NOT localized
	if(UnitFactionGroup("player") == "Alliance") then
		if(IsEquippedItem(47914) or IsEquippedItem(47984) or IsEquippedItem(48035))		then n = 1; end
		if(IsEquippedItem(47981) or IsEquippedItem(47987) or IsEquippedItem(48029))		then n = n + 1; end
		if(IsEquippedItem(47936) or IsEquippedItem(47986) or IsEquippedItem(48031)) 	then n = n + 1; end
		if(IsEquippedItem(47982) or IsEquippedItem(47983) or IsEquippedItem(48037)) 	then n = n + 1; end
		if(IsEquippedItem(47980) or IsEquippedItem(47985) or IsEquippedItem(48033)) 	then n = n + 1; end
	else
		if(IsEquippedItem(48068) or IsEquippedItem(48065) or IsEquippedItem(48058)) 	then n = 1; end
		if(IsEquippedItem(48071) or IsEquippedItem(48062) or IsEquippedItem(48061))		then n = n + 1; end
		if(IsEquippedItem(48070) or IsEquippedItem(48063) or IsEquippedItem(48060)) 	then n = n + 1; end
		if(IsEquippedItem(48067) or IsEquippedItem(48066) or IsEquippedItem(48057)) 	then n = n + 1; end
		if(IsEquippedItem(48069) or IsEquippedItem(48064) or IsEquippedItem(48059)) 	then n = n + 1; end	
	end
	
	if(n >= 4) then
		privateScaling["4pcRaid9"] = 1;
	else
		privateScaling["4pcRaid9"] = 0;
	end
end

local function priest_OnLevelUp()
	priest_UpdatePlayerScaling();
end

local function priest_OnTalentUpdate()
	priest_ScanTalents();
	priest_UpdatePlayerScaling();
end

local function priest_OnEquipmentChangedDelayed()
	priest_ScanEquipment();	
	priest_UpdatePlayerScaling();
end

local function priest_OnEquipmentChanged()
	AM_Core:ScheduleUniqueTimer("priest_equip", priest_OnEquipmentChangedDelayed, 0.7);
end

local function priest_OnScalingEncode()
	return {UnitLevel("player"), privateScaling.base, privateScaling.sp, privateScaling.DA};
end

AM_Public.OnScalingDecode.PRIEST = function(guid, in_guidScaling)	
	if(#in_guidScaling ~= 4) then return; end

	priest_ApplyScaling(guid, unpack(in_guidScaling));
end

AM_Public.OnEnableClass.PRIEST = function()
	AM_Public.OnLevelUp = priest_OnLevelUp;
	AM_Public.OnTalentUpdate = priest_OnTalentUpdate;
	AM_Public.OnEquipmentChanged = priest_OnEquipmentChanged;	
	AM_Public.OnScalingEncode = priest_OnScalingEncode;
	
	priest_ScanTalents();
	priest_ScanEquipment();
		
	priest_UpdatePlayerScaling();
end


-- #################################################################################################


-- Public Scaling: { [AstralShift] }
-- We default to 0.3, because quite frankly nobody will use less points except while leveling
local shaman_defaultScaling = {0.3};


local function shaman_AstralShift_Create(sourceGUID, sourceName, destGUID, destName, spellId, destEffects)
	local sourceScaling, quality = AM_Unit_Scaling(sourceGUID, shaman_defaultScaling, 0.7);
	
	return -1, quality, sourceScaling[1];
end

local function shaman_AstralShift_Hit(guid, effectEntry, absorbedRemaining, overkill, spellSchool)
	local total = absorbedRemaining + overkill;

	return floor(total * effectEntry[7]), true;
end

local function shaman_OnTalentUpdate()
	-- Astral Shift
	local _, _, _, _, t = GetTalentInfo(1, 21);
	
	playerScaling[1] = t * 0.1;
	
	AM_Public.ScheduleScalingBroadcast();
end

AM_Public.OnEnableClass.SHAMAN = function()
	AM_Public.OnTalentUpdate = shaman_OnTalentUpdate;
		
	shaman_OnTalentUpdate();
end


-- #################################################################################################


-- TODO: base leveling increase
local warlock_Sacrifice_Spells = {
	[7812] = 305,
	[19438] = 510,
	[19440] = 770,
	[19441] = 1095,
	[19442] = 1470,
	[19443] = 1905,
	[27273] = 2855,
	[47985] = 6750,
	--[47986] = 8350,
	-- There is a new rank of Sacrifice at 79, so it scales by 15 to 80
	[47986] = 8365,
}

local warlock_ShadowWard_Spells = {
	[6229] = 290,
	[11739] = 470,
	[11740] = 675,
	[28610] = 875,
	[47890] = 2750,
	[47891] = 3300,
};

-- Public Scaling: { [DemonicBrutality] }
local warlock_defaultScaling = {1.0};


-- No downranking support here
local function warlock_Sacrifice_Create(sourceGUID, sourceName, destGUID, destName, spellId, destEffects)
	-- Note that the source is the voidwalker, so dest is the warlock!
	local sourceScaling, quality = AM_Unit_Scaling(destGUID, warlock_defaultScaling, 0.4);

	return floor(warlock_Sacrifice_Spells[spellId] * sourceScaling[1]), quality;
end

local function warlock_ShadowWard_Hit(guid, effectEntry, absorbedRemaining, overkill, spellSchool)
	if(spellSchool ~= SCHOOL_MASK_SHADOW) then
		return 0, true;
	end
	
	return generic_Hit(guid, effectEntry, absorbedRemaining, overkill, spellSchool);
end

local function warlock_OnTalentUpdate()
	-- Demonic Brutality
	local _, _, _, _, t = GetTalentInfo(2, 6);
	
	playerScaling[1] = 1 + (t * 0.1);
	
	AM_Public.ScheduleScalingBroadcast();	
end

AM_Public.OnEnableClass.WARLOCK = function()
	AM_Public.OnTalentUpdate = warlock_OnTalentUpdate;
	
	warlock_OnTalentUpdate();
end


-- #################################################################################################


local function items_EssenceOfGossamer_Hit(guid, effectEntry)
	if(effectEntry[3] < 140) then
		return effectEntry[3], false;
	else
		return 140, true;
	end
end

local function items_ArgussianCompass_Hit(guid, effectEntry)
	if(effectEntry[3] < 68) then
		return effectEntry[3], false;
	else
		return 68, true;
	end
end

local function items_Valanyr_OnAuraApplied(sourceGUID, sourceName, destGUID, destName, spellId)
	AM_Public.AddCombatTrigger(sourceGUID, "OnHeal", items_Valanyr_OnHeal);
end

local function items_Valanyr_OnAuraRemoved(sourceGUID, sourceName, destGUID, destName, spellId)
	AM_Public.RemoveCombatTrigger(sourceGUID, "OnHeal", items_Valanyr_OnHeal);
end

local function items_Valanyr_Create(sourceGUID, sourceName, destGUID, destName, spellId, destEffects)
	local existing = 0;
	
	if(destEffects and destEffects[spellId]) then
		existing = destEffects[spellId][3];
	end

	local charge = AM_PopCharge(destGUID, spellId);
	
	if(charge == 0) then
		return existing, 0.0;
	end
		
	-- According to the blue post explaining the Val'anyr effect on introduction, all units are
	-- contributing to the same bubble with a cap of 20.000
	return min(20000, existing + charge), 1.0;
end

local function items_Valanyr_OnHeal(sourceGUID, sourceName, destGUID, destName, spellId, amount)
	AM_PushCharge(destGUID, 64413, floor(amount * 0.15), 5.0);
end

local function items_Stoicism_Create(sourceGUID, sourceName, destGUID, destName, spellId, destEffects)
	local maxHealth = UnitHealthMax(destGUID, destName);
	
	if(not maxHealth) then
		return 0, 0.0;
	end
	
	return floor(maxHealth * 0.2), 1.0;	
end


-- #################################################################################################


local mage_FireWard_Entry = {2.0, 30, EffectType.BUFF, generic_SpellScalingByTable_Create, mage_FireWard_Hit, mage_Absorb_Spells, 0.8053};
local mage_FrostWard_Entry = {2.0, 30, EffectType.BUFF, generic_SpellScalingByTable_Create, mage_FrostWard_Hit, mage_Absorb_Spells, 0.8053};
local mage_IceBarrier_Entry = {1.0, 60, EffectType.BUFF, mage_IceBarrier_Create, generic_Hit};
local mage_ManaShield_Entry = {1.0, 60, EffectType.BUFF, generic_SpellScalingByTable_Create, generic_Hit, mage_Absorb_Spells, 0.8053};

local priest_PWS_Entry = {1.0, 30, EffectType.BUFF, priest_PowerWordShield_Create, generic_Hit};	

local warlock_Sacrifice_Entry = {1.0, 30, EffectType.BUFF, generic_ConstantByTable_Create, generic_Hit, warlock_Sacrifice_Spells};
local warlock_ShadowWard_Entry = {2.0, 30, EffectType.BUFF, generic_SpellScalingByTable_Create, warlock_ShadowWard_Hit, warlock_ShadowWard_Spells, 0.8053};


-- INCOMPLETE
-- This table gets a lot smaller in Cataclysm
AM_Public.Effects = {
	-- Unknown Effect
	-- This is used when a known effect is applied, but it is impossible to properly account for it,
	-- for example if an AREA effect is applied with an unknown trigger
	[0] = {1.0, 0, EffectType.BUFF, function() return 0, 0.0; end, nil};
	
	-- DEATH KNIGHT	
	-- Anti-Magic Shell
	[48707] = {3.0, 5, EffectType.BUFF, deathknight_AntiMagicShell_Create, deathknight_AntiMagicShell_Hit},
	-- Anti-Magic Zone
	[50461] = {3.0, 10, EffectType.AREA, deathknight_AntiMagicZone_Create, deathknight_AntiMagicZone_Hit},
	-- Will of the Necropolis
	[49189] = {0.0, 0, EffectType.PASSIVE, nil, deathknight_WillOfTheNecropolis_Hit},
		
	-- DRUID
	-- Savage Defense
	[62606] = {1.1, 10, EffectType.BUFF, druid_SavageDefense_Create, druid_SavageDefense_Hit},
	
	-- MAGE
	-- Fire Ward
	[543] = mage_FireWard_Entry,
	[8457] = mage_FireWard_Entry,
	[8458] = mage_FireWard_Entry,
	[10223] = mage_FireWard_Entry,
	[10225] = mage_FireWard_Entry,
	[27218] = mage_FireWard_Entry,
	[43010] = mage_FireWard_Entry,
	-- Frost Ward
	[6143] = mage_FrostWard_Entry,
	[8461] = mage_FrostWard_Entry,
	[8462] = mage_FrostWard_Entry,
	[10177] = mage_FrostWard_Entry,
	[28609] = mage_FrostWard_Entry,
	[32796] = mage_FrostWard_Entry,
	[43012] = mage_FrostWard_Entry,
	-- Ice Barrier
	[11426] = mage_IceBarrier_Entry,
	[13031] = mage_IceBarrier_Entry,
	[13032] = mage_IceBarrier_Entry,
	[13033] = mage_IceBarrier_Entry,
	[27134] = mage_IceBarrier_Entry,
	[33405] = mage_IceBarrier_Entry,
	[43038] = mage_IceBarrier_Entry,
	[43039] = mage_IceBarrier_Entry,			
	-- Mana Shield
	[1463] = mage_ManaShield_Entry,
	[8494] = mage_ManaShield_Entry,
	[8495] = mage_ManaShield_Entry,
	[10191] = mage_ManaShield_Entry,
	[10192] = mage_ManaShield_Entry,
	[10193] = mage_ManaShield_Entry,
	[27131] = mage_ManaShield_Entry,
	[43019] = mage_ManaShield_Entry,
	[43020] = mage_ManaShield_Entry,		
	
	-- PALADIN
	-- Sacred Shield
	[58597] = {1.0, 6, EffectType.BUFF, paladin_SacredShield_Create, generic_Hit},
	-- Ardent Defender
	
	-- PRIEST	
	-- Power Word: Shield
	[17] 	= priest_PWS_Entry,
	[592]	= priest_PWS_Entry,
	[600] 	= priest_PWS_Entry,
	[3747] 	= priest_PWS_Entry,
	[6065] 	= priest_PWS_Entry,
	[10898] = priest_PWS_Entry,
	[10899] = priest_PWS_Entry,
	[10900] = priest_PWS_Entry,
	[10901] = priest_PWS_Entry,
	[25217] = priest_PWS_Entry,
	[25218] = priest_PWS_Entry,
	[48065] = priest_PWS_Entry,
	[48066] = priest_PWS_Entry,
	-- Divine Aegis
	[47753] = {1.0, 12, EffectType.BUFF, priest_DivineAegis_Create, generic_Hit},
	
	-- ROGUE
	-- Cheat Death
	--   No clue how it works technically (does it still uses an absorb?)
	--   While there is no absorb value to watch for, it would screw our
	--   numbers on other effects.
	
	-- SHAMAN
	-- Astral Shift
	[52179] = {2.5, nil, EffectType.BUFF, shaman_AstralShift_Create, shaman_AstralShift_Hit},
	-- Glyph of Stoneclaw Totem
	--   This absorb effect cannot be supported without heavy exceptional
	--   handling, because it is neither applied by SPELL_AURA_APPLIED or
	--   nor removed by SPELL_AURA_REMOVED - but only an UNIT_AURA event is
	--   fired on gain/loss. The gain can actually be detected by the cast 
	--   of Stoneclaw Totem (and knowledge about the glyphs), but not if it
	--   breaks prematurely, making looping through all buffs needed everytime
	--	 it could be active and UNIT_AURA is fired. At least the amount would
	--	 be constant for every rank.
	--   The normal totem absorb effect could be tracked with more ease, but still
	--   there is no proper removal event, so it would rely on zombified absorb
	--   effects until the totem drops. Maybe later. (What about totem GUIDs?)
	-- [55277] = {1.0, shaman_GlyphStoneclawTotem_Create, generic_Hit},
	
	-- WARLOCK
	-- Sacrifice
	[7812] = warlock_Sacrifice_Entry,
	[19438] = warlock_Sacrifice_Entry,
	[19440] = warlock_Sacrifice_Entry,
	[19441] = warlock_Sacrifice_Entry,
	[19442] = warlock_Sacrifice_Entry,
	[19443] = warlock_Sacrifice_Entry,
	[27273] = warlock_Sacrifice_Entry,
	[47985] = warlock_Sacrifice_Entry,
	[47986] = warlock_Sacrifice_Entry,
	-- Shadow Ward
	[6229] = warlock_ShadowWard_Entry,
	[11739] = warlock_ShadowWard_Entry,
	[11740] = warlock_ShadowWard_Entry,
	[28610] = warlock_ShadowWard_Entry,
	[47890] = warlock_ShadowWard_Entry,
	[47891] = warlock_ShadowWard_Entry,

	-- ITEMS	
	-- Val'anyr (spellId of the created absorb effect)
	[64413] = {1.0, 8, EffectType.BUFF, items_Valanyr_Create, generic_Hit},
	-- Essence of Gossamer
	[60218] = {5.0, 10, EffectType.BUFF, function() return 4000, 1.0; end, items_EssenceOfGossamer_Hit},
	-- Corroded Skeleton Key
	[71586] = {1.0, 10, EffectType.BUFF, function() return 6400, 1.0; end, generic_Hit},
	-- Phaseshift Bulwark
	[36481] = {1.0, 4, EffectType.BUFF, function() return 100000, 1.0; end, generic_Hit},
	-- Darkmoon Card: Illusion
	[57350] = {1.0, 6, EffectType.BUFF, function() return 1500, 1.0; end, generic_Hit},
	-- Mark of the Dragon Lord
	[17252] = {1.0, 1800, EffectType.BUFF, function() return 500, 1.0; end, generic_Hit},
	-- The Burrower's Shell
	[29506] = {1.0, 20, EffectType.BUFF, function() return 900, 1.0; end, generic_Hit},
	-- Arena Grand Master (mean value of 1000)
	[23506] = {1.0, 20, EffectType.BUFF, function() return 1000, 0.5; end, generic_Hit},
	-- Argussian Compass
	[39228] = {5.0, 20, EffectType.BUFFL, function() return 1150, 1.0; end, items_ArgussianCompass_Hit},
	-- Runed Fungalcap
	[31771] = {1.0, 20, EffectType.BUFF, function() return 440, 1.0; end, generic_Hit},
	-- Truesilver Champion
	[9800] = {1.0, 60, EffectType.BUFF, function() return 175, 1.0; end, generic_Hit},
	-- Gnomish Harm Prevention Belt
	[13234] = {1.0, 600, EffectType.BUFF, function() return 500, 1.0; end, generic_Hit},
	-- Nigh Invulnerability Belt
	[30458] = {1.0, 8, EffectType.BUFF, function() return 4000, 1.0; end, generic_Hit},
	-- Divine Protection (Priest Dungeon Set 1/2 4pc bonus)
	[27779] = {1.0, 30, EffectType.BUFF, function() return 350, 1.0; end, generic_Hit},
	-- Armor of Faith (Priest Raid Set 3 4pc bonus)
	[28810] = {1.0, 30, EffectType.BUFF, function() return 500, 1.0; end, generic_Hit},
	-- Lesser Ward of Shielding
	[29674] = {1.0, nil, EffectType.BUFF, function() return 1000, 1.0; end, generic_Hit},
	-- Greater Ward of Shielding
	[29719] = {1.0, nil, EffectType.BUFF, function() return 4000, 1.0; end, generic_Hit},
	-- Stoicism (Warrior Raid Set 10 4pc bonus)
	[70845] = {1.0, 10, EffectType.BUFF, items_Stoicism_Create, generic_Hit},
	
	
	-- Scarab Brooch
	--   This trinket is a bit more problematic, because it creates non-stacking
	--   shields, having always the strongest up. The question is now, what happens
	--   if a previously stronger shield gets hit and reduced below a might-be-created
	--   shield coming from a following heal? It also seems like only a certain
	--   set of healing spells attribute to bubbles
};


AM_Public.AreaTriggers = {
	-- Anti-Magic Zone
	[51052] = 50461,
	
	-- Power Word: Barrier
	[62618] = 81782,
};


AM_Public.CombatTriggers = {	
	-- Val'anyr
	["64411_OnAuraApplied"] = items_Valanyr_OnAuraApplied,
	["64411_OnAuraRemoved"] = items_Valanyr_OnAuraRemoved,
};
